/*
   Copyright 2008-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.util;

import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;

/**
 * An abstract variable that just has get/set.
 * This is usefule when a final variable is needed, but the contents need to be changed.
 * This also uses the Observable/Observer pattern to allow dependent logic to react to value changes.
 *
 */
public class ValueHolder<T> extends Observable {

    //------------------------------------------------------------
    //- Class Variables


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private T fNullDefault = null;
    private T fValue;
    private final Map<ValueChangeHandler<T>, Adapter<T>> fHandlers = new IdentityHashMap<ValueChangeHandler<T>, Adapter<T>>();


    //------------------------------------------------------------
    //- Constructors

    public ValueHolder(T pValue) {
        set(pValue);
    }


    //------------------------------------------------------------
    //- Getters

    public T get() {
        return fValue;
    }


    //------------------------------------------------------------
    //- Settors

    /**
     * Set the new value, if it is different (determined by equals()) than the current value.
     *
     * @param pValue  The candidate new value.  If it is null and a NullDefault is defined, then that value is set.
     */
    public void set(T pValue) {
        // First check if the new value is null and should be defaulted to an object value.
        T newValue = (pValue == null) ? fNullDefault : pValue;

        boolean valueChanged;
        if (fValue == null) {
            valueChanged = (newValue != null);
        } else {
            valueChanged = !fValue.equals(newValue);
        }

        if (valueChanged) {
            fValue = newValue;

            setChanged();
            notifyObservers(newValue);
        }
    }

    /**
     * Define how to handle 'null' values.  This value is substituted for 'null'.
     *
     * @param pNullDefault The object value to substitue for 'null'.
     * @return  this for cascading.
     */
    public ValueHolder<T> setNullDefault(T pNullDefault) {
        fNullDefault = pNullDefault;

        return this;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions


    //------------------------------------------------------------
    //- Public Interface Functions

    public ValueHolder<T> addHandler(ValueChangeHandler<T> pHandler) {
        Adapter<T> adapter = new Adapter<T>(pHandler);
        // Do not allow duplicate handlers.
        if (!fHandlers.containsKey(pHandler)) {
            fHandlers.put(pHandler, adapter);
            addObserver(adapter);
            pHandler.handleInitialValue(get());
        }

        return this;
    }

    public ValueHolder<T> removeHandler(ValueChangeHandler<T> pHandler) {
        Adapter<T> adapter = fHandlers.remove(pHandler);
        if (adapter != null) {
            deleteObserver(adapter);
        }

        return this;
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    @Override
    public String toString() {
        return "ValueHolder[" + get() + "]";
    }


    //------------------------------------------------------------
    //- Inner Classes

    private class Adapter<T> implements Observer {
        private final ValueChangeHandler<T> fHandler;

        private Adapter(ValueChangeHandler<T> pHandler) {
            fHandler = pHandler;
        }

        public void update(Observable pObservable, Object pObject) {
            //noinspection unchecked
            fHandler.handleValueChange((T) pObject);
        }
    }


    //------------------------------------------------------------
    //- Main

}
